function [bestX, bestFx] = exscout(fx, XLow, XHi, InitStep, FinStep, MaxIters, ...
                                 MaxPop, MaxScoutPop, MaxScoutPop2 , ...
                                 bSkipNarrowTrusRegion, bTrace)
% The SCOUT function implements a version of the Fireworks
% Algorithm.
% This version uses an additional/secondary set of scouts.
%
% INPUT
% =====
% fx - handle of optimized function.
% XLow - array of lower limits of trust region.
% XHi - array of upper limits of trust region.
% MaxIters - maximum number of iterations.
% MaxPop - the population size.
% MaxScoutPop - the size of the local scout population.
% MaxScoutPop2 - the size of the secondary local scout population.
% bSkipNarrowTrusRegion - flag used to narrow the trus range before
% performing the optimization calculation.
% bTrace - optional Boolean flag used to trace intermediate results.
%
% OUTPUT
% ======
% bestX - array of best solution.
% bestFx - best function value.
%
% Example
% =======
% >> [bestX, bestFx] = exscout(@fx2, [0 0 0 0], [5 5 5 5], 0.1, 0.0001, 1000, 100, 100, 50, true)
%

if nargin < 11, bTrace = false; end
if nargin < 10, bSkipNarrowTrusRegion = false; end
tpc = fix(MaxIters/10); % calculate then % of total iterations
n = length(XLow);
if length(XHi) < n, XHi = XHi(1) + zeros(1,n); end
m = n + 1;

pop = zeros(MaxPop,m);
scoutPop = zeros(MaxScoutPop,m);
scoutPop2 = zeros(MaxScoutPop2,m);

if ~bSkipNarrowTrusRegion
  [XLow, XHi, pop] = RangeZoom(fx, XLow, XHi, MaxIters, MaxPop, 0.1, 3);
  fprintf('New trust region is\n');
  fprintf('Xlow -> ')
  fprintf('%f ', XLow);
  fprintf('\nXHi -> ');
  fprintf('%f ', XHi);
  fprintf('\n');
else
  pop = zeros(MaxPop,m);
  pop(:,m) = 1e+99;
  for i=1:MaxPop
    pop(i,1:n) = XLow + (XHi - XLow) .* rand(1,n);
    pop(i,m) = fx(pop(i,1:n));
  end
  pop = sortrows(pop,m);
end

if bTrace
  fprintf('%f ', pop(1,1:n));
  fprintf('%e\n', pop(1,m));
end

for iter = 1:MaxIters
  StepFactor = exp(log(InitStep) + log(FinStep/InitStep)*log(iter)/log(MaxIters));
  
  % update main population, whenevr possible
  popMember = zeros(1,m);
  for i=2:MaxPop
    if rand > 0.5
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = pop(i,1:n) + (pop(1,1:n) - pop(i,1:n)) .* rand(1,n) * sf;
    else
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = XLow + (XHi - XLow) .* rand(1,n) * sf;
    end
    popMember(m) = fx(popMember(1:n));
    if popMember(m) < pop(i,m)
      pop(i,:) = popMember;
    end
  end
  pop = sortrows(pop,m);  
  
  for i=1:MaxPop
    m2 = fix(MaxScoutPop/2);
    % Use uniform random perturbations
    for j=1:m2
      for k=1:n
        scoutPop(j,k)  = pop(i,k) + (XHi(k) - XLow(k)) * StepFactor * (2*rand-1);
        if scoutPop(j,k) < XLow(k) || scoutPop(j,k) > XHi(k)
          scoutPop(j,k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end      
      end
      scoutPop(j,m) = fx(scoutPop(j,1:n));
      % ---------------------- Seconday scouts
      m3 = fix(MaxScoutPop2/2);
      % Use uniform random perturbations
      for j2=1:m3
        for k2=1:n
          scoutPop2(j2,k2)  = scoutPop(j,k2) + sqrt(XHi(k2) - XLow(k2)) * StepFactor * (2*rand-1);
          if scoutPop2(j2,k2) < XLow(k2) || scoutPop2(j2,k2) > XHi(k2)
            scoutPop2(j2,k2) = XLow(k2) + (XHi(k2) - XLow(k2)) * rand;
          end      
        end
        scoutPop2(j2,m) = fx(scoutPop2(j2,1:n));
      end
      % Use normal random perturbations
      for j2=m3+1:MaxScoutPop2
        for k2=1:n
          scoutPop2(j2,k2) = normrand(scoutPop(j,k2), sqrt((XHi(k2) - XLow(k2))/3) * StepFactor);
          if scoutPop2(j2,k2) < XLow(k2) || scoutPop2(j2,k2) > XHi(k2)
            scoutPop2(j2,k2) = XLow(k2) + (XHi(k2) - XLow(k2)) * rand;
          end         
        end
        scoutPop2(j2,m) = fx(scoutPop2(j2,1:n));
      end      
      
      scoutPop2 = sortrows(scoutPop2,m);
      scoutPop(j,:) =  scoutPop2(1,:);
    end %for j
    % Use normal random perturbations
    for j=m2+1:MaxScoutPop
      for k=1:n
        scoutPop(j,k) = normrand(pop(i,k), (XHi(k) - XLow(k)/3) * StepFactor);
        if scoutPop(j,k) < XLow(k) || scoutPop(j,k) > XHi(k)
          scoutPop(j,k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end    
      end
      scoutPop(j,m) = fx(scoutPop(j,1:n));
      
      % ---------------------- Seconday scouts
      m3 = fix(MaxScoutPop2/2);
      % Use uniform random perturbations
      for j2=1:m3
        for k2=1:n
          scoutPop2(j2,k2)  = scoutPop(j,k2) + sqrt(XHi(k2) - XLow(k2)) * StepFactor * (2*rand-1);
          if scoutPop2(j2,k2) < XLow(k2) || scoutPop2(j2,k2) > XHi(k2)
            scoutPop2(j2,k2) = XLow(k2) + (XHi(k2) - XLow(k2)) * rand;
          end      
        end
        scoutPop2(j2,m) = fx(scoutPop2(j2,1:n));
      end
      % Use non-uniform random perturbations
      for j2=m3+1:MaxScoutPop2
        for k2=1:n
          scoutPop2(j2,k2) = normrand(scoutPop(j,k2), sqrt((XHi(k2) - XLow(k2))/3) * StepFactor);
          if scoutPop2(j2,k2) < XLow(k2) || scoutPop2(j2,k2) > XHi(k2)
            scoutPop2(j2,k2) = XLow(k2) + (XHi(k2) - XLow(k2)) * rand;
          end         
        end
        scoutPop2(j2,m) = fx(scoutPop2(j2,1:n));
      end      
      
      scoutPop2 = sortrows(scoutPop2,m);
      scoutPop(j,:) =  scoutPop2(1,:);      
    end % for j =
    pop = [pop; scoutPop];
    pop = sortrows(pop,m);
    pop = pop(1:MaxPop,:);
  end
  
  if bTrace && mod(iter, tpc) == 0
    fprintf('%f ', pop(1,1:n));
    fprintf('%20.18e, step_factor=%e\n', pop(1,m), StepFactor);
  end
end

bestX = pop(1,1:n);
bestFx = pop(1,m);
end

function y = normrand(mean,sigma)
  y = mean + sigma * randn(1,1);
end