function [bestX, bestFx] = exscout3(fx, XLow, XHi, InitStep, FinStep, MaxIters,...
                                 MaxPop, MaxScoutPop, bSemiLogDecay, ...
                                 bSkipNarrowTrusRegion, bTrace)
% The SCOUT function implements an extended version of SOA.
% This version uses both uniform and exponential ratio RNG to perform perturbation on
% the scout local search. The function uses a single loop to determine the 
% best way to calculate a scout location (using uniform and exponential ratio random
% numbers)
%
% INPUT
% =====
% fx - handle of optimized function.
% XLow - array of lower limits of trust region.
% XHi - array of upper limits of trust region.
% InitStep - initial search step size.
% FinStep - final search step size.
% MaxIters - maximum number of iterations.
% MaxPop - the population size.
% MaxScoutPop - the size of the local scout population.
% bSemiLogDecay - flag used to use log-linear step decay (when true) or
% log-log step decayscout (when false).
% bSkipNarrowTrusRegion - flag used to narrow the trus range before
% performing the optimization calculation.
% bTrace - optional Boolean flag used to trace intermediate results.
%
% OUTPUT
% ======
% bestX - array of best solution.
% bestFx - best function value.
%
% Example
% =======
%

if nargin < 11, bTrace = false; end
if nargin < 10, bSkipNarrowTrusRegion = false; end
if nargin < 9, bSemiLogDecay = true; end
if nargin < 8, MaxScoutPop = fix(MaxPop/4); end
tpc = fix(MaxIters/10); % calculate then % of total iterations
n = length(XLow);
if length(XHi) < n, XHi = XHi(1) + zeros(1,n); end
m = n + 1;
minR = exp(0);
maxR = exp(1);

%---------------------- Narrow Trust Region ----------------
if ~bSkipNarrowTrusRegion
  [XLow, XHi, pop] = RangeZoom(fx, XLow, XHi, MaxIters, MaxPop, 0.1, 3);
  fprintf('New trust region is\n');
  fprintf('Xlow -> ')
  fprintf('%f ', XLow);
  fprintf('\nXHi -> ');
  fprintf('%f ', XHi);
  fprintf('\n');
else
  pop = zeros(MaxPop,m);
  pop(:,m) = 1e+99;
  for i=1:MaxPop
    pop(i,1:n) = XLow + (XHi - XLow) .* rand(1,n);
    pop(i,m) = fx(pop(i,1:n));
  end
  pop = sortrows(pop,m);
end

%------------------------------------------------------------
scoutPop = zeros(MaxScoutPop,m);
scoutPop(:,m) = 1e+99;
scout1 = zeros(1,m);
scout2 = zeros(1,m);
if bTrace
  fprintf('%f ', pop(1,1:n));
  fprintf('%e\n', pop(1,m));
end

for iter = 1:MaxIters
  % StepFactor = InitStep + (FinStep - InitStep)*(iter-1)/(MaxIters-1);
  if bSemiLogDecay
    StepFactor = exp(log(InitStep) + (iter-1)/(MaxIters-1) * log(FinStep/InitStep));
  else
     StepFactor = exp(log(InitStep) + log(FinStep/InitStep)*log(iter)/log(MaxIters));
  end
  
  % update main population, whenevr possible
  popMember = zeros(1,m);
  for i=2:MaxPop
    if rand > 0.5
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = pop(i,1:n) + (pop(1,1:n) - pop(i,1:n)) .* rand(1,n) * sf;
    else
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = XLow + (XHi - XLow) .* rand(1,n) * sf;
    end
    popMember(m) = fx(popMember(1:n));
    if popMember(m) < pop(i,m)
      pop(i,:) = popMember;
    end
  end
  pop = sortrows(pop,m);
  
  for i=1:MaxPop
    for j=1:MaxScoutPop
      % Use uniform random perturbations
      for k=1:n
        scout1(k)  = pop(i,k) + (XHi(k) - XLow(k)) * StepFactor * (2*rand-1);
        if scout1(k) < XLow(k) || scout1(k) > XHi(k)
          scout1(k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end        
      end
      scout1(m) = fx(scout1(1:n));

      % Use non-uniform random perturbations
      for k=1:n
        scout2(k) = exporatiorand(pop(i,k), (XHi(k) - XLow(k))/3 * StepFactor, minR, maxR);
        if scout2(k) < XLow(k) || scout2(k) > XHi(k)
          scout2(k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end        
      end
      scout2(m) = fx(scout2(1:n));
      
      % select the better scout
      if scout1(m) < scout2(m)
        scoutPop(j,:) = scout1;
      else
        scoutPop(j,:) = scout2;
      end
    end
    pop = [pop; scoutPop];
    pop = sortrows(pop,m);
    pop = pop(1:MaxPop,:);
  end
  
  if bTrace && mod(iter, tpc) == 0
    fprintf('%f ', pop(1,1:n));
    fprintf('%20.18e, step_factor=%e\n', pop(1,m), StepFactor);
  end
end

bestX = pop(1,1:n);
bestFx = pop(1,m);
end

function y = exporatiorand(mean, sigma, minR, maxR)
 
  r1 = rand;
  r2 = rand;
  x = exp(r1)/exp(r2);
  x = (x - minR)/(maxR - minR);
  
  y = mean + sigma * x;
  
  y = (y + 0.3674) / (0.9964 + 0.3674);
  y = 2*y - 1;
end