function [bestX, bestFx] = scout6(fx, XLow, XHi, InitStep, FinStep, MaxIters,...
                                 MaxPop, MaxScoutPop, bSemiLogDecay, ...
                                 bSkipNarrowTrusRegion, bTrace)
% The SCOUT function implements a version of the Fireworks
% Algorithm. This version uses both uniform and a second cosine RNG to perform 
% perturbation on the scout local search.
%
% INPUT
% =====
% fx - handle of optimized function.
% XLow - array of lower limits of trust region.
% XHi - array of upper limits of trust region.
% InitStep - initial search step size.
% FinStep - final search step size.
% MaxIters - maximum number of iterations.
% MaxPop - the population size.
% MaxScoutPop - the size of the local scout population.
% bSemiLogDecay - flag used to use log-linear step decay (when true) or
% log-log step decayscout (when false).
% bSkipNarrowTrusRegion - flag used to narrow the trus range before
% performing the optimization calculation.
% bTrace - optional Boolean flag used to trace intermediate results.
%
% OUTPUT
% ======
% bestX - array of best solution.
% bestFx - best function value.
%
% Example
% =======
%

if nargin < 11, bTrace = false; end
if nargin < 10, bSkipNarrowTrusRegion = false; end
if nargin < 9, bSemiLogDecay = true; end
if nargin < 8, MaxScoutPop = fix(MaxPop/4); end
tpc = fix(MaxIters/10); % calculate then % of total iterations
n = length(XLow);
if length(XHi) < n, XHi = XHi(1) + zeros(1,n); end
m = n + 1;

%---------------------- Narrow Trust Region ----------------
if ~bSkipNarrowTrusRegion
  [XLow, XHi, pop] = RangeZoom(fx, XLow, XHi, MaxIters, MaxPop, 0.1, 3);
  fprintf('New trust region is\n');
  fprintf('Xlow -> ')
  fprintf('%f ', XLow);
  fprintf('\nXHi -> ');
  fprintf('%f ', XHi);
  fprintf('\n');
else
  pop = zeros(MaxPop,m);
  pop(:,m) = 1e+99;
  for i=1:MaxPop
    pop(i,1:n) = XLow + (XHi - XLow) .* rand(1,n);
    pop(i,m) = fx(pop(i,1:n));
  end
  pop = sortrows(pop,m);
end

%------------------------------------------------------------
scoutPop = zeros(MaxScoutPop,m);
scoutPop(:,m) = 1e+99;
if bTrace
  fprintf('%f ', pop(1,1:n));
  fprintf('%e\n', pop(1,m));
end

for iter = 1:MaxIters
  % StepFactor = InitStep + (FinStep - InitStep)*(iter-1)/(MaxIters-1);
  if bSemiLogDecay
    StepFactor = exp(log(InitStep) + (iter-1)/(MaxIters-1) * log(FinStep/InitStep));
  else
     StepFactor = exp(log(InitStep) + log(FinStep/InitStep)*log(iter)/log(MaxIters));
  end
  
  % update main population, whenever possible
  popMember = zeros(1,m);
  for i=2:MaxPop
    if rand > 0.5
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = pop(i,1:n) + (pop(1,1:n) - pop(i,1:n)) .* rand(1,n) * sf;
    else
      sf = StepFactor;
      if rand > 0.5, sf = 1; end
      popMember(1:n) = XLow + (XHi - XLow) .* rand(1,n) * sf;
    end
    popMember(m) = fx(popMember(1:n));
    if popMember(m) < pop(i,m)
      pop(i,:) = popMember;
    end
  end
  pop = sortrows(pop,m); 
  
  for i=1:MaxPop
    m2 = fix(MaxScoutPop/2);
    % Use uniform random perturbations
    for j=1:m2
      for k=1:n
        scoutPop(j,k)  = pop(i,k) + (XHi(k) - XLow(k)) * StepFactor * (2*rand-1);
        if scoutPop(j,k) < XLow(k) || scoutPop(j,k) > XHi(k)
          scoutPop(j,k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end
        
      end
      scoutPop(j,m) = fx(scoutPop(j,1:n));
    end
    % Use non-uniform random perturbations
    for j=m2+1:MaxScoutPop
      for k=1:n
        scoutPop(j,k) = cosrand2(pop(i,k), (XHi(k) - XLow(k))/3 * StepFactor);
        if scoutPop(j,k) < XLow(k) || scoutPop(j,k) > XHi(k)
          scoutPop(j,k) = XLow(k) + (XHi(k) - XLow(k)) * rand;
        end
        
      end
      scoutPop(j,m) = fx(scoutPop(j,1:n));
    end
    pop = [pop; scoutPop];
    pop = sortrows(pop,m);
    pop = pop(1:MaxPop,:);
  end
  
  if bTrace && mod(iter, tpc) == 0
    fprintf('%f ', pop(1,1:n));
    fprintf('%20.18e, step_factor=%e\n', pop(1,m), StepFactor);
  end
end

bestX = pop(1,1:n);
bestFx = pop(1,m);
end

function y = cosrand2(mean,sigma)
  
  r = rand;
  x = (r - mean)/sigma;
  y = exp(-x^2)*cos(x);
  chs = sign(y);
  chs2 = chs;
  if chs==0, chs2 = 1; end
  y = mean + chs * sigma / (chs2 + abs(log10(abs(y))));  
  y = (y - 0.5877)/(1 - 0.5887);
  y = 2*y - 1;
end